//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
//  Program reversing the contracting operation. See Section 5. of the paper.
//
//  Executing: 
//
// cat tmp_kminimal_faces | sed 's/[^ ]*:[^ ]*//g' >  "tmp_in_kminimal_faces"
// cat tmp_k6_faces       | sed 's/[^ ]*:[^ ]*//g' >  "tmp_in_k6_faces"
// ./cycle_glue  > "out_6-critical_nonisomorf"
//
 
#include <iostream>
#include <fstream>
#include <vector>
#include <queue>

#define PRINT_TRY_ERROR 1

using namespace std;

const char* C_OUT_DRAWINGS = "out_6-critical_drawings";
const char* C_IN_K_MINIMAL = "tmp_in_kminimal_faces";
const char* C_IN_K6        = "tmp_in_k6_faces";



const int MAX_VERTICES = 50;
const int MAX_FACES    = 100;
const int MAX_K        = 12;

int tested = 0;
int unique = 0;

typedef vector<int>  face; // index of vertices going clockwise

struct graph
{
	int           vertices;
	int           k;   // k-minimality. -1 if the graph is not minimal at all :)
	
	vector<face> faces; 	
	
	bool                  neigh_valid;
	vector< vector<int> >   neighbours;
	
	template<class T>
	void print_neighbours(T &stream) const
	{
		for (int i = 0; i < vertices; i++)
		{
			stream << "v" <<  i << ": " << neighbours[i].size() << "   ";
			for (unsigned int j = 0; j < neighbours[i].size(); j++)
			{
				stream << neighbours[i][j] << " ";
			}
			stream << endl;
		}
	}
	
	template<class T>
	void print_faces(T &stream) const
	{
		for (unsigned int i = 0; i < faces.size(); i++)
		{
			stream << "f" << i << ": " << faces[i].size() << "   ";
			for (unsigned int j = 0; j < faces[i].size(); j++)
			{
				stream << faces[i][j] << " ";
			}
			stream << endl;
		}		
	}
	
	
	// Only vertex representation
	void add_edge(int vertexA, int vertexB)
	{
		// Security chceck
		for (unsigned int i = 0; i < neighbours[vertexA].size(); i++)
			if (neighbours[vertexA][i] == vertexB) 
			{
				
				cerr << "BUG_A " << vertexA << " " << vertexB << endl;
				print_neighbours(cerr);
				exit (1);
			}

		for (unsigned int i = 0; i < neighbours[vertexB].size(); i++)
			if (neighbours[vertexB][i] == vertexA) cerr << "BUG_B" << endl;

		neighbours[vertexA].push_back(vertexB);
		neighbours[vertexB].push_back(vertexA);
	}
	
	int add_vertex()
	{
		//neighbours.resize(vertices+1);
		neigh_valid = false;
		return vertices++;
	}
	
	void set_v_f(int v, int f)
	{
		neighbours.resize(v);
		faces.resize(f);
		vertices = v;
		neigh_valid = false;
	}
	
	void clear_vert_rep()
	{
		neigh_valid = false;
		neighbours.clear();
	}
	
	// removes edge only in vertex representation
	void remove_edge(int vertexA, int vertexB)
	{
		// from vertex A
		for (int i = 0; i < (signed int)neighbours[vertexA].size(); i++)
		{
			if (neighbours[vertexA][i] == vertexB)
			{
				if ( i != (signed int)neighbours[vertexA].size() - 1)
				{
					neighbours[vertexA][i] = neighbours[vertexA][neighbours[vertexA].size() - 1];
				}
				neighbours[vertexA].pop_back();
			}
		}

		// from vertex B
		for (int i = 0; i < (signed int)neighbours[vertexB].size(); i++)
		{
			if (neighbours[vertexB][i] == vertexA)
			{
				if ( i != (signed int)neighbours[vertexB].size() - 1)
				{
					neighbours[vertexB][i] = neighbours[vertexB][neighbours[vertexB].size() - 1];
				}
				neighbours[vertexB].pop_back();
			}
		}
	}
	
	void erase()
	{
		neighbours.clear();
		vertices = 0;
		neigh_valid = false;
		faces.clear();
	}
	
	graph()
	{
			k = -1;
			vertices = 0;
			neigh_valid = false;
	}
};
 
// --------------------------------------------------------------------------


template<class T>
T& operator<<(T &stream, const face &f)
{
	stream << "[";
	for (unsigned int i = 0; i < f.size(); i++)
	{
		stream << f[i] << " ";
	}
	stream << "]";
	return stream;
}


vector<graph> kminimal[MAX_K];
vector<graph> tested_graphs_before_split;
vector<graph> tested_graphs_for_criticality;
vector<graph> unique_graphs;
queue<graph>  to_process;

//// Update 
void create_vert_rep(graph &g);


void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}

template<class T>
int at(const vector<T> &vec, int idx)
{
	int sz = vec.size();
	return vec[(idx+sz)%sz];
}


template<class T>
T& operator<<(T &stream, const graph &g)
{
	stream << g.k << " v: " << g.vertices << " f: " << g.faces.size() << endl;
	g.print_faces(stream);
	g.print_neighbours(stream);	
	stream << endl;
	
	return stream;
}

template<class T>
void print_dan_compatible(T &stream, const graph & g)
{
	stream << g.k << " v: " << g.vertices << " f: " << g.faces.size() << endl;
	
	// faces
	g.print_faces(stream);
	stream << endl;
}

template<class T>
void print_vertices(T &stream, const graph & g)
{
	stream << g.k << " v: " << g.vertices << " f: " << g.faces.size() << endl;
	
	// faces
	g.print_neighbours(stream);
	
	stream << endl;
}



template<class T>
bool load_dan_compatible(T &stream, graph & g)
{
	g.erase();
		
	stream >> g.k;
	if (!stream) return false;
	
	
	int vertices;
	stream >> vertices;
	if (!stream) return false;

	int faces;
	stream >> faces;
	if (!stream) return false;

	g.set_v_f(vertices, faces);
	
	for (int i = 0; i < faces; i++)
	{
		int fsz;
		stream >> fsz;
		
		if (!stream)
		{
			cerr << "Something wrong while loading face " << i << endl;
			return false;
		}

		int last  = -1;
		int first = -1;
		int e;
		
		for (int j = 0; j < fsz; j++)
		{
			stream >> e;

			if (first == -1)
			{
				last = first = e;
			}
			else
			{
				last = e;
			}
			
			if (!stream)
			{
				cerr << "Something wrong with loading vertex " << j << " at face " << i << endl;
				return false;
			}
			
			g.faces[i].push_back(e);
		}
		
	}
	
	return true;
}


template<class T>
bool load_neighbours_list(T &stream, graph & g)
{
	g.erase();
	
	stream >> g.k;
	
	if (!stream) return false;
	
	int vertices;
	stream >> vertices;
	
	if (!stream) return false;
	
	g.set_v_f(vertices, 0);
	
	for (int i = 0; i < vertices; i++)
	{
		int degree;
		stream >> degree;
		
		if (!stream)
		{
			cerr << "Something wrong while loading vertex " << i << endl;
			return false;
		}
		
		
		for (int j = 0; j < degree; j++)
		{
			int e;
			stream >> e;
			
			if (!stream)
			{
				cerr << "Something wrong with loading neighbour " << j << " at neighbour " << i << endl;
				return false;
			}
			
			if (e == i)
			{
				cerr << "Loop at vertex " << i << endl;
				return false;
			}
			
			if (e > i)
			{
				//	cerr << e << "-" << i << endl;
				g.add_edge(e, i);
			}
		}
	}
	
	g.neigh_valid = true;
	
	return true;
}


bool check_validity(graph &g, bool print_reason = false)
{
	int edges[g.vertices][g.vertices];
	
	bzero(edges, sizeof(int)*g.vertices*g.vertices);
	
	for (unsigned int i = 0; i < g.faces.size(); i++)
	{
		for (unsigned int j = 0; j < g.faces[i].size(); j++)
		{
			int a = g.faces[i][j];
			int b = at(g.faces[i], j+1);
			if (a > b) swap(a,b);
			if (a == b) return false;
			
			edges[a][b]++;
		}
	}
	
	for (int i = 0; i < g.vertices; i++)
	{
		for (int j = i+1; j < g.vertices; j++)
		{
			if (edges[i][j] != 0 && edges[i][j] != 2) 
			{
				if (print_reason)
				{
					cerr << "edge invalid: " << i << ", " << j << " in graph "<< endl;
					cerr << g;
				}
				return false;
			}
		}
	}
	
	return true;
}

/////////////////////////////////////////////////////////////// Isomorphism testing

// Return false if they have different degrees/vertices
bool check_degrees(graph &a, graph &b)
{
	if (a.vertices != b.vertices)
	{
		return false;
	}

	create_vert_rep(a);
	create_vert_rep(b);
		
	for (int deg = 1; deg < a.vertices - 1; deg++)
	{
		int cnta = 0;
		for (int i = 0; i < a.vertices; i++)
			if ((signed int)a.neighbours[i].size() == deg) cnta++;
		
		int cntb = 0;
		for (int i = 0; i < b.vertices; i++)
			if ((signed int)b.neighbours[i].size() == deg) cntb++;
		
		if (deg == 1 && (cnta != 0 || cntb !=0))
		{
			cerr << "Vertices of degree 1? " << endl;
			cerr << a << b;
			exit(1);
		}
		
		if (cnta != cntb) return false;
	}
	
	return true;
}


int map_atob[MAX_VERTICES];
int map_btoa[MAX_VERTICES];


int map_face_btoa[MAX_FACES];

bool try_map_face(const graph &a, const graph &b, unsigned int face_in_a)
{	
	if (face_in_a >= a.faces.size()) return true;
	
	unsigned int sz = a.faces[face_in_a].size();

	
	for (unsigned int i = 0; i < b.faces.size(); i++)
	{
		
		if (map_face_btoa[i] != 0) continue;
		
		if (b.faces[i].size() != sz) continue;
		
		for (int reverse = 0; reverse < 2; reverse++)
		{
		
			for (unsigned int shift = 0; shift < sz; shift++)
			{
				unsigned int chckID;
				for (chckID = 0; chckID < sz; chckID++)
				{
					if (reverse)
					{
						if (b.faces[i][chckID] != map_atob[ a.faces[face_in_a][(-chckID+shift+sz)%sz] ]) break;
					}
					else
					{
						if (b.faces[i][chckID] != map_atob[ a.faces[face_in_a][(chckID+shift)%sz] ]) break;
					}
				}
				
				if (chckID >= sz)
				{
					map_face_btoa[i] = 1;
					if (try_map_face(a, b, face_in_a+1)) return true;
					map_face_btoa[i] = 0;
				}			
			}
		}
	}
	return false;
}

bool is_mapping_ok_for_cycles(const graph &a, const graph &b)
{
	if (a.faces.size() != b.faces.size()) return false;
	
	
	// try to check sizes of faces
	for (int deg = 1; deg < a.vertices; deg++)
	{
		int cnta = 0;
		for (unsigned int i = 0; i < a.faces.size(); i++)
			if ((signed int)a.faces[i].size() == deg) cnta++;
		
		int cntb = 0;
		for (unsigned int i = 0; i < b.faces.size(); i++)
			if ((signed int)b.faces[i].size() == deg) cntb++;
		
		if (cnta != cntb) return false;
	}
	
	// Ok we have same number of faces and they have same sizes.
	// So try to map the faces.
		
	bzero(map_face_btoa, MAX_FACES*sizeof(int));
	
	return try_map_face(a, b, 0);
	
}

bool can_map(const graph &a, const graph &b, int i, int j)
{
	if (map_btoa[j] != -1) return false;
	if (a.neighbours[i].size() !=  b.neighbours[j].size()) return false;
//	cerr << "sz= " << a.neighbours[i].size() << "==" <<  b.neighbours[j].size() << endl;
	
	// test if all previous neighbours are mapped correctly
	for (unsigned int vertA = 0; vertA < a.neighbours[i].size(); vertA++)
	{
//		cerr << "vertA == " << vertA << " i == " << i << " sz = " << a.neighbours[i].size() << endl;
		if (a.neighbours[i][vertA] > i) continue;
		int mapped = map_atob[a.neighbours[i][vertA]];
		
		unsigned int vertB;
		for (vertB = 0; vertB < b.neighbours[j].size(); vertB++)
			if (b.neighbours[j][vertB] == mapped) break;
		
		// we found previously mapped vertex adjacent to i but it's
		// image is not mapped as a neigbbour of j.
		if (vertB == b.neighbours[j].size()) return false;
	}
	
	return true;
}

bool match_othervertices(const int i, const graph &a, const graph &b, bool match_faces)
{	
	if (i >= a.vertices) 
	{
//		cerr << "vertices matched" << endl;
//		return (!match_faces || is_mapping_ok_for_cycles(a, b));
		if (!match_faces || is_mapping_ok_for_cycles(a, b))
		{
			//cerr << "mapping b -> a is: " << endl;
			//for (int j = 0; j < b.vertices; j++)
			//	cerr << j << "->" << map_btoa[j] << endl;
			return true;
		}
		else
		{
			return false;
		}
	}
	
	// scroll of the list of first vertex
	for (int j = 0; j < a.vertices; j++)
	{
		if (!can_map(a, b, i, j)) continue;
		
		// fill the mapping
		map_atob[i] = j;
		map_btoa[j] = i;
		
		if (match_othervertices(i+1, a, b, match_faces)) return true;
		
		map_btoa[j] = -1;
	}
	
	return false;
}

bool are_isomorphic(graph &a, graph &b, bool match_faces)
{
	// try map a vertex
	//A -> B

	create_vert_rep(a);
	create_vert_rep(b);	

	if (!check_degrees(a, b)) 
	{
		//		cerr << "quick test failed" << endl;
		return false;
	}	
	
	for (int i = 0; i < a.vertices; i++)
	{
		map_atob[i] = -1;	
		map_btoa[i] = -1;	
	}
	
	if (match_faces)
	{
//		cerr << "Matching....." << endl;
//		cerr << a << b;
//		print_dan_compatible(cerr, a);
//		print_dan_compatible(cerr, b);		
	}
	
	return match_othervertices(0, a, b, match_faces);
}

/////////////////////////////////////////////////////////////// Coloring testing


int coloring[MAX_VERTICES];


// Check if vertex color is not confligting with other
// previous vertices. true == all is allright
bool check_vertex(const graph &g, const int vertexID)
{
//	cerr << "v: " << vertexID << " " << g.neighbours[vertexID].size();
	for (unsigned int testedID = 0; testedID < g.neighbours[vertexID].size(); testedID++)
	{
//		cerr << g.neighbours[vertexID][testedID] << " ";
		if (coloring[vertexID] == coloring[g.neighbours[vertexID][testedID]]
			&& g.neighbours[vertexID][testedID] < vertexID) 
		{
			return false;
		}
	}
//	cerr << endl;
	
	return true;
}


bool precolor_vertex(const graph &g, const int i)
{
	if (i >= g.vertices) 
	{
		
//		for (int j = 0; j < i; j++)
//		{
//			cerr << coloring[j] << " ";
//		}
//		cerr << endl;
//		cerr << "this is strange" << endl;
		
		return true;
	}
		
	// Use always color 0 on vertex 0.
	if ( i == 0 )
	{
		coloring[0] = 0;
		
		// recursive
		return precolor_vertex(g, 1);
	}
	else
	{
		for (int col = 0; col < 5; col++)
		{	
			coloring[i] = col;
				
			// some chord...
			if (!check_vertex(g, i)) continue; 
				
			// recursive
			if (precolor_vertex(g, i+1)) return true;
		}
		return false;
	}
}


bool try_graph_after_split(graph &g)
{
	
	if (tested++ % 50 == 0) 
	{
		cerr << tested << ":" << tested_graphs_before_split.size() << endl;
	}
	
	for (unsigned int i = 0; i < tested_graphs_before_split.size(); i++)
	{
		if (are_isomorphic(tested_graphs_before_split[i], g, true)) return false;
	}
	tested_graphs_before_split.push_back(g);
	
	return true;
}
	

bool is_6_critical(graph &g)
{
	create_vert_rep(g);
	
	// 5-pre coloring test - if all precolorings are
	// extensible we are not interested in such a graph
	// but it could be extended...
	if (precolor_vertex(g, 0)) 
	{	
//		cerr << "Precoloring failed" << endl;		
		return false;
	}
	
	
	// Test if it is l-minimalcritical
	for (int v = 0; v < g.vertices; v++)
	{
		vector<int>   neighbours = g.neighbours[v]; // we must do the copy
		
		for (unsigned int i = 0; i < neighbours.size(); i++)
		{
			if ( neighbours[i] < v )
			{
				g.remove_edge(v, neighbours[i]);
				
				bool rv = precolor_vertex(g, 0);
				
			//	print_vertices(cerr, g);
				
				g.add_edge(v, neighbours[i]);
				
				if (!rv)
				{
	//				cerr << "Falese witness - not critical beacuse of edge " << v << " " << neighbours[i] << endl;
					return false;
				}
			}
		}
	}
	
	return true;
}

// Try to add new graph
void try_graph(graph &g, int v3)
{
//	if (tested++ % 5000 == 0) 
//	{
//		cerr << tested << ":" << tested_graphs.size() << endl;
//	}
	
	create_vert_rep(g);	
	if (g.neighbours[v3].size() != 5)
	{
#ifdef	PRINT_TRY_ERROR		
//		cerr << "Splitted vertex has not degree 5" << endl;
// Then the splitting is invalid and we may skip this case
#endif				
		return;	
	}
	
	if (!is_6_critical(g)) return;
	
	
	// Test if it is not isomorphic to other graphs
	for (unsigned int gID = 0; gID < unique_graphs.size(); gID++)
	{
		if (are_isomorphic(g, unique_graphs[gID], true)) 
		{
			//cerr << "Isomorf one" << endl;
			return;
		}
	}
	
	unique_graphs.push_back(g);
	to_process.push(g);
		
//	cout << "Cool - we have a new graph :-)" << endl;
//	cout << "G:" << unique_graphs.size() << " " << g;

	return;
}

/////////////////////////////////////////////////////////////// Faces subdivision

bool glue(graph &g, int faceID, const graph &to_glue, const int rotation, const bool flipping)
{
	
//	cerr << "----- Gluing ----- " << rotation << " " << flipping << endl;
//	print_dan_compatible(cerr, g);
//	print_dan_compatible(cerr, to_glue);
	
	if (to_glue.k !=  (signed int)g.faces[faceID].size())
	{
		cerr << "INVALID SIZES FOR GLUE!!!" << endl;
		return false;
	}
	
	const face &f = g.faces[faceID];
	
//	cerr << "face: " << f << endl;
	
	// Create first part of mapping
	int mapping[to_glue.vertices];
	for (int i = 0; i < to_glue.k; i++)
	{
		if (!flipping) mapping[i] = f[(i+rotation)%to_glue.k];
		else mapping[i] = f[(to_glue.k-i+rotation)%to_glue.k];
	}
	
	// remove old face
	if (faceID != (signed int)g.faces.size()-1)
	{
		g.faces[faceID] = g.faces[g.faces.size()-1];
	}
	g.faces.pop_back(); // erase the old face
	
	// complete mapping (add new vertices
	for (int i = to_glue.k; i < to_glue.vertices; i++)
	{
		mapping[i] = g.add_vertex();
	}
			
	// add faces from to_glue
	for (unsigned int i = 0; i < to_glue.faces.size(); i++)
	{
		face f;
		
		for (unsigned int j = 0; j < to_glue.faces[i].size(); j++)
		{
			f.push_back(mapping[to_glue.faces[i][j]]);
		}
				
		g.faces.push_back(f);
	}
	
	g.clear_vert_rep();
/*	
	cerr << "result: " << endl;
	cerr << "mapping: " << endl;
	for (int i = 0; i < to_glue.vertices; i++)
	{
		cerr << mapping[i] << " ";
	}
	
	cerr << endl;
	
	print_dan_compatible(cerr, g);
*/
	return check_validity(g);
}

bool find_first(const graph &g, int v, int faceID, int &previous, int &next)
{
	for (unsigned int i = 0; i < g.faces[faceID].size(); i++)
	{
			if (g.faces[faceID][i] == v)
			{
				next     = at(g.faces[faceID], i+1);
				previous = at(g.faces[faceID], i-1);
			}
	}
	return false;
}

int find_next(const graph &g, int v, int faceID, int &previous, int &next, int &extraold)
{
	for (unsigned int i = 0; i < g.faces[faceID].size(); i++)
	{
		if (g.faces[faceID][i] == previous)
		{
			if ( at(g.faces[faceID], i+1) == v && at(g.faces[faceID], i+2) != extraold)
			{
				next = at(g.faces[faceID], i+2);
				return true;
			}
		}
	}
	return false;
}

int find_next_backward(const graph &g, int v, int faceID, int &previous, int &next, int &extraold)
{
	for (unsigned int i = 0; i < g.faces[faceID].size(); i++)
	{
		if (g.faces[faceID][i] == v && at(g.faces[faceID], i+1) == previous
			&& at(g.faces[faceID], i-1) != extraold)
			{
				next = at(g.faces[faceID], i-1);
				return true;
			}
	}
	return false;
}

bool cycle_order(graph &g, int v, vector<int> &faces, vector<int> &vertices)
{
	// Tu se udela optimalizace, aby se braly jen 
	// steny, kde se v vyskytuje
	
	faces.clear();
	vertices.clear();
	
	int first_previous = -1;
	int first_next     = -1;
	int i;
	for (i = 0; i < (signed int)g.faces.size(); i++)
	{
		if (find_first(g, v, i, first_previous, first_next)) break;
	}
	
	if (first_previous == -1)
	{
		cerr << "Vertex " << v << " is not in any face" << endl;
		return false;
	}
	
	int previous = first_next;
	int next = -1;
	int extraold = -1;
	
//	cerr << "[" << first_previous << "," << v << "," << first_next << "] " ;
	
	int xt = g.vertices;
	
	do
	{
		int faceID;
		for (faceID = 0; faceID < (signed int)g.faces.size(); faceID++)
		{
			if (find_next(g, v, faceID, previous, next, extraold)) break;
		}
		
		if (faceID == (signed int)g.faces.size())
		{
			// try reverse search
			for (faceID = 0; faceID < (signed int)g.faces.size(); faceID++)
			{
				if (find_next_backward(g, v, faceID, previous, next, extraold)) break;
			}
			
			if (faceID == (signed int)g.faces.size())
			{
				//cerr << "Bugy bugy vertex: " << v << " previous: " << previous << " extra old: " << extraold << endl;
				
				// the only possibility is that we have a vertex of degree 2
				vertices.clear();
				vertices.push_back(first_previous);
				vertices.push_back(first_next);
				
				break;
				//print_dan_compatible(cerr, g);
				//return false;
			}
		}
		
		
		
//		cerr << " " << faceID << "[" << previous << "," << v << "," << next << "] " ;
		
		faces.push_back(faceID);
		vertices.push_back(previous);
		
		extraold = previous;
		previous = next;
	} while ((next != first_next || extraold != first_previous) && --xt > -1);
	
//	cerr << endl;
	
	if (xt < 0)
	{
		cerr << "FUJJ: " << v << endl;
		cerr  << "[" << previous << "," << v << "," << next << "] " ;
		print_dan_compatible(cerr, g);
		return false;
	}
	
	return true;
}


// Creating vertex representation from faces representation
void create_vert_rep(graph &g)
{
//	if (g.neigh_valid) return;
	
	if (!check_validity(g, true))
	{
		cerr << "Invalid graph for creating vert rep!!!" << endl;
	}

//	cerr << "creating vert rep" << endl;
	
	vector<int> tmp;
	g.neighbours.resize(g.vertices);
	for (int v = 0; v < g.vertices; v++)
	{
		if (!cycle_order(g, v, tmp, g.neighbours[v]))
		{
			cerr << "Something wrong in cyclic ordering" << endl;
			exit(1);
		}
		if (g.neighbours[v].size() < 2)
		{
			cerr << "Vertex of degree 1? in create_vert_rep " << endl;
			exit(0);
		}
	}
	
	g.neigh_valid = true;
}

// Fill the last face with all k-minimal
// performs all rotations and flipping of the minimal graph
void glue_faces_from_list(graph &h, vector<int> faces_to_glue, int v3)
{
	if (faces_to_glue.size() == 0) 
	{
		try_graph(h, v3);
		return;
	}
	
	int faceID = faces_to_glue[faces_to_glue.size()-1];
	faces_to_glue.pop_back();

	int fsz = h.faces[faceID].size();
	
	// try not gluing anytrhing to the face
	glue_faces_from_list(h, faces_to_glue, v3);
	
	// try gluing something
	for (unsigned int glueID = 0; glueID < kminimal[fsz].size(); glueID++)
	{
		for (unsigned int rotation = 0; rotation < h.faces[faceID].size(); rotation++)
		{
			for (int flip = 0; flip < 2; flip++)
			{
				graph tmp = h;
				if (!glue(tmp, faceID, kminimal[fsz][glueID], rotation, flip)) continue;
				
				//try_graph(tmp, v3);
				
				glue_faces_from_list(tmp, faces_to_glue, v3);
				
			}
		}
	}
}

// Initialize the gluing procedure - create list of all faces of size
// > 3. Observe that there is nothing to glue to faces of size 3.
void glue_to_all_faces(graph &h, int v3)
{
	// here could be an isomorphism test

	vector<int> faces_to_glue;

	for (unsigned int f = 0; f < h.faces.size(); f++)
	{
		if (h.faces[f].size() > 3) faces_to_glue.push_back(f);
		if (h.faces[f].size() > 10)
		{
			cerr << "Found face of size > 10!!" << endl;
		}
	}

	glue_faces_from_list(h, faces_to_glue, v3);


}

//////////////////////////////////////////////////////////// Splitting

void repaint(graph &g, int faceID, int previous, int current, int next, int splitted)
{
	face &f = g.faces[faceID];
	for (int i = 0; i < (signed int)f.size(); i++)
	{
		if (f[i] == current && (   (at(f, i-1) == previous &&  at(f, i+1) == next) 
			                     ||(at(f, i-1) == next     &&  at(f, i+1) == previous)) )
		{
			f[i] = splitted;
			return;
		}
	}

	cerr << "Not repainting!!!" << endl;
	cerr << "repainting face" << faceID << ": " << previous << " " << current << " " <<  next << " -> " << splitted << endl;

}

/////////////
/// previsou current next  -> previous repl1 repl2 repl3  next
/// next current previous  -> next     repl3 repl2 repl1  prevuois

void insert_instead_current(graph &g, int faceID, int previous, int current, int next, int repl1, int repl2, int repl3)
{	
	face f;
	const face &gf =  g.faces[faceID];
	
	for (int i = 0; i < (signed int)gf.size(); i++)
	{
		if (gf[i] == current)
		{
			if (at(gf, i-1) == previous && at(gf, i+1) == next)
			{
				f.push_back(repl1);
				f.push_back(repl2);
				f.push_back(repl3);
				continue;
			}
			if (at(gf, i-1) == next && at(gf, i+1) == previous)
			{
				f.push_back(repl3);
				f.push_back(repl2);
				f.push_back(repl1);
				continue;
			}
		}
		f.push_back(gf[i]);
	}

	if (f.size() != gf.size()+2)
	{
		cerr << "buggy insert_instead_current: " << faceID << " "  << previous << " " << current << " " << next 
		     << " to " << repl1 << " " << repl2 << " " << repl3<< endl;
		print_dan_compatible(cerr, g);
		exit(1);
	}
	
//	cerr << g.faces[faceID] << " -> " << f <<  "  (" << previous << " " << current << " " << next << ")" 
//		<< " ->  (" << repl1 << " " << repl2 << " " << repl3 << ")" << endl;
	
	g.faces[faceID] = f;
}


void split_vertices(graph &g)
{
	vector<int> faces;
	vector<int> vertices;
	
	
//	5 == 9;
	for (int v = 0; v < g.vertices; v++)
	{
		cycle_order(g, v, faces, vertices);
			
		for (unsigned int i = 0; i < faces.size(); i++)
		{
			for (unsigned int j = i+1; j < faces.size(); j++)
			{
			
// TESTING STUFF //
				if (g.faces[faces[i]].size() == 3) continue;
				if (g.faces[faces[j]].size() == 3) continue;
				
// splitting in the 5 face				
				
				graph h = g;
				
				int v2 = h.add_vertex(); // outer part of splitted vertex
				int v3 = h.add_vertex(); // inner splitted vertex [must have deg==5 after glue]
	
//				cerr << "----------------- " << i << " " << j << endl;
//				cerr << "splittting " << v << " in faces " << faces[i] << " " << faces[j] << " and vertices " << v2 << " " << v3 << endl;
//				print_dan_compatible(cerr, h);
			
	//			cerr << "vertices added" << endl;
				
				// repaint
				for (unsigned int k = i + 1; k < j; k++)
				{
					repaint(h, faces[k], vertices[k], v, at(vertices, k+1), v2);
				}

//				cerr << "Repained" << endl;
	
				
				// split
				insert_instead_current(h, faces[i], vertices[i], v, at(vertices, i+1), v, v3, v2);
				insert_instead_current(h, faces[j], vertices[j], v, at(vertices, j+1), v2, v3, v);
				
				// OK the graph h is ready for pasting in faces[i] and faces[j]
				
//				cerr << "splitted " << v << " in faces " << faces[i] << " " << faces[j] << endl;

				if (!try_graph_after_split(h)) 
				{
				//	cerr << "allready found" << endl;
					continue;
				}
			
				
				// there will be some gluing and copying....
				h.clear_vert_rep();
				
				glue_to_all_faces(h, v3);
				
			}
		}
	}
}

/////////////////////////////////////////////////////////////// Main

template<class T>
void load_kminimal(T &stream)
{
	 graph g;
	 
	 while(load_dan_compatible(stream, g))
	 {
		 if (g.k < 0 || g.k > 10)
		 {
			 cerr << "Invalid k in kminimal. Must be between 0 and 10." << endl;
			 exit(1);
		 }
		 kminimal[g.k].push_back(g);
	 }
}

template<class T>
void load_toprocess(T &stream)
{
	graph g;
	
	int processed = 0;
	
	while(load_dan_compatible(stream, g))
	{
		if (g.k != -1)
		{
			cerr << "Invalid L in to process - must be -1 - at graph " << processed << endl;
			exit(1);
		}
		
		if (!is_6_critical(g)) 
		{
			cerr << "Not critical graph: " << processed << endl;
			exit(1);
		};
		
		
		// Test if it is not isomorphic to other graphs
		unsigned int gID;
		for (gID = 0; gID < unique_graphs.size(); gID++)
		{
			if (are_isomorphic(g, unique_graphs[gID], true)) 
			{
				cerr << "Graph " <<  processed << " isomorf to allready loaded " << gID << endl;
				exit(1);
			}
		}
		if (gID < unique_graphs.size()) continue;
		
		
		unique_graphs.push_back(g);
		to_process.push(g);
		processed++;		
	}
}


template<class T>
void write_drawings(T &stream)
{
	// write out drawings
	for (unsigned int gID = 0; gID < unique_graphs.size(); gID++)
	{
		stream << "G:" << gID << " " << unique_graphs[gID];
	}
	
	cerr << "Found " << unique_graphs.size() << " drawings" << endl;
}

int main(int argc, char **argv)
{
	
	// The kminimal for gluing
	ifstream kmin_s(C_IN_K_MINIMAL);
	load_kminimal(kmin_s);

	// For processing
	ifstream toprocess_s(C_IN_K6);
	load_toprocess(toprocess_s);

	cerr << "input loaded" << endl;
	
	
///////////// For testing	

//	graph g, h;
//    load_dan_compatible(cin, g);
  //  load_dan_compatible(cin, h);

	
	
//	create_vert_rep(g);
	
//	print_vertices(cerr, g);

//return (0);
	
//	glue(g, 1, to_glue, 1, false);
	
//	print_dan_compatible(cerr, g);
	
//	cerr << "--splitting--" << endl;
	
//	split_vertices(g);

	// Do the splitting

	while (!to_process.empty())
	{
		split_vertices(to_process.front());
		to_process.pop();
		
		cerr << "Queue size: " << to_process.size() << endl;
		
	}
	
	// Writing out the result
		
	ofstream out_faces_s(C_OUT_DRAWINGS);
	write_drawings(out_faces_s);

	vector<graph> uniq_results;
	
	for (unsigned int gID = 0; gID < unique_graphs.size(); gID++)
	{
		unsigned int i;
		for (i = 0; i < uniq_results.size(); i++)
		{
			if (are_isomorphic(uniq_results[i], unique_graphs[gID], false)) break;
		}
		if (i != uniq_results.size()) continue;
		
		uniq_results.push_back(unique_graphs[gID]);
		cout << "G:" << i << " " << unique_graphs[gID];
	}
	
	cerr << "Done, found " << uniq_results.size() << " 6-critical graphs." << endl;
}
