//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
//  Program for generating all embedings of K_6.
//  Only the embeddings to the klein bottle are printed.
//  The embeddings are printed to stdout, info messages are
//  printed to stderr. See Section 4. for detail description
//
//  Executing: 
//
//  ./k6_generator > tmp_k6_unique
//


#include <iostream>

using namespace std;


int   neighbours[6][5]; // List of neighbours
bool  edgesflips[6][6]; // nighbour matrix of edgeflips


// Swap integers a and b
void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}

long long searched = 0; // So far searched graphs
int found = 0;  // so far found graphs

// Create a next permutation of array. It is assumed
// that array has size 5. Generaly it is expected to be
// called as next_perm(neighbours[i]);
void next_perm(int *array)
{
	// Find first nonincreasing from back
	int first = 4;
	while (first > 1 && array[first] < array[first-1]) first--;
	if (first > 1) first--;
	
	// Place the next larger to the front
	int nextmin = first + 1;
	for (int i = first + 1; i < 5; i++)
	{
		if (array[nextmin] > array[i] && array[i] > array[first]) nextmin = i;
	}
	swap(array[nextmin], array[first]);
	
	// sort the rest
	for (int i = first+1; i < 5; i++)
		for (int j = first+1; j < 4; j++)
		{
			if(array[j] > array[j+1]) swap(array[j], array[j+1]);
		}
}

// Init a permutation of 5 vertices.
// assumed to be called as init_perm(neighbours[i], i);
// result  neigbrbours[i] = {1, 2,.. i-1, i+1, .. 5 }
void init_perm(int *array, int skip)
{
	int k = 0;
	for (int j = 0; j < 5; j++)
	{
		if (skip == k) k++;
		array[j] = k++;
	}
}

// Print the neighbours matrix - for debugging
void print_neigh()
{
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 5; j++)
		{
			cout << neighbours[i][j] << " ";
		}
		cout << endl;
	}
}

// print the edge flips matrix - for debugging
void print_edgeflips()
{
	// init neighbours
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			cout << edgesflips[i][j] << " ";
		}
		cout << endl;
	}	
}

// Used edges are for enumerating faces. We enumare them
// by walking along the cycles and check if we use an edge.
// possible valies are  
//   0  ... not used
//   1  ... one side used
//   2  ... second side used
//   3  ... both sides used
int  used_edges[6][6];

// Start walking and keep walking along an edge. We want to go
// from vertex i to vertex j and use edgeside side of the edge.
// the matrix used_edges is filled with face containing this
// edge (assuming that no edges from the face are marked)
void walk_edge(int i, int j, int edgeside)
{
//	cout << "i=" << i << ", j=" << j;
//	cout << i << "("<< j << "," << edgeside << ");";
	
	// check if the edge was allready visited. If so just return.
	if (used_edges[i][j] & edgeside) return;
	
	used_edges[i][j] |= edgeside;
	
	// flip the edge is applicable
	if (edgesflips[i][j])
	{
			if (edgeside == 1) edgeside = 2;
			else edgeside = 1;
	}

	// makr destination of edge
	if (edgeside == 1) used_edges[j][i] |= 2;
	else used_edges[j][i] |= 1;

	// find next vertex in the face
	int idx = 0;
	while (neighbours[j][idx] != i) idx++;
	
	if (edgeside == 2)
	{
		if (++idx > 4) idx = 0;
	}
	else
	{
	 	if (--idx < 0) idx = 4;
	}
		
	walk_edge(j, neighbours[j][idx], edgeside);
}

// For debugging purposes printing an used edges
void print_used_edges()
{
	// init neighbours
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			cout << used_edges[i][j] << " ";
		}
		cout << endl;
	}	
}

// Print a current graph to an output stream
template<class T>
void print_kral_compatible(int genus, T &stream)
{
	// vertices
	for (int i = 0; i < 6; i++)
		stream << i << ":" <<  neighbours[i][0] <<  neighbours[i][1] <<  neighbours[i][2]
			   <<  neighbours[i][3] <<  neighbours[i][4] << " ";
	
	// adjacency matrix
	for (int i = 0; i < 5; i++) 
	{
		for (int j = i+1; j < 6; j++)
			stream << edgesflips[i][j];
		stream << " ";
	}
	stream << "  G" << genus << '\n';
}

// Test if we have a wanted graph. It means we need
// to check the genus of the grap
void check_graph_validity()
{
	// Try to count faces.
	bzero(used_edges, sizeof(int)*36);

//	print_used_edges();
	
	int faces = 0;
	
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			if (i == j) continue;
			
			if ((used_edges[i][j] & 1) == 0)
			{
				walk_edge(i,j,1);
				faces++;
			}
			
			if ((used_edges[i][j] & 2) == 0)
			{
				walk_edge(i,j,2);
				faces++;
			}			
		}
	}
	
//	cout << faces << endl;

	
	// Try if genus == 2
	int genus = 11 - faces;
	
	//cout << genus << endl;
	
	if (genus == 2)
	{
		found++;
		print_kral_compatible(genus, cout);
	}
	
	if ((++searched % 100000000) == 0)
	{
		cerr << "status: " << found <<  "/" << searched <<  endl;
		print_kral_compatible(genus, cerr);
	}
	
}

// Generate all possible edge flips and try if the resulting graph
// is one of the wanted graphs.
void generate_edgesflips()
{
	// we skip the option where all edges are non-inverted.
	for (int i = 1; i < (2 << 9); i++)
	{
		edgesflips[5][4] = edgesflips[4][5] = i &    1;
		edgesflips[5][3] = edgesflips[3][5] = i &    2;
		edgesflips[4][3] = edgesflips[3][4] = i &    4;
		edgesflips[5][2] = edgesflips[2][5] = i &    8;
		edgesflips[4][2] = edgesflips[2][4] = i &   16;
		edgesflips[3][2] = edgesflips[2][3] = i &   32;
		edgesflips[5][1] = edgesflips[1][5] = i &   64;
		edgesflips[4][1] = edgesflips[1][4] = i &  128;
		edgesflips[3][1] = edgesflips[1][3] = i &  256;
		edgesflips[2][1] = edgesflips[1][2] = i &  512;

		check_graph_validity();
	}
}

// Generate permutations for all but the firs vertex.
// We can allways rename the vertices such that the
// list of first vertex is {1,2,3,4,5}. Thus we only
// need to generate the rest.
void generate_vertexlists()
{
	for (int i = 0; i < 24; i++)
	{
		for (int i = 0; i < 24; i++)
		{
			for (int i = 0; i < 24; i++)
			{
				for (int i = 0; i < 24; i++)
				{
					for (int i = 0; i < 24; i++)
					{
						
						//	print_neigh();
						generate_edgesflips();
						//cout << endl;
						
						next_perm(neighbours[5]);
					}
					init_perm(neighbours[5],5);						
					next_perm(neighbours[4]);
				}
				init_perm(neighbours[4],4);					
				next_perm(neighbours[3]);
			}
			init_perm(neighbours[3],3);
			next_perm(neighbours[2]);
		}
		init_perm(neighbours[2],2);
		next_perm(neighbours[1]);
	}
}


int main(int argc, char **argv)
{
	// init neighbours
	for (int i = 0; i < 6; i++)
	{
		int k = 0;
		for (int j = 0; j < 5; j++)
		{
			if (i == k) k++;
			neighbours[i][j] = k++;
		}
	}

	// reset the diagonal of edge flips
	for (int i = 0; i < 6; i++)
	{
		edgesflips[i][i] = 0;
	}
	

	// generate the all
	generate_vertexlists();
	
	return 0;
}


