//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
// Program for testing isomorfisms of drawings of K6.
// on stdin the program reads a drawings of k6 to the projective
// plane and output only nonisomorf drawings.
//
//  Executing: 
//
//  ./k6_generator | sed 's/.://g' | sed 's/G//' | ./k6_isomorfer
//  
//  cat tmp_k6 | sed 's/.://g' | sed 's/G//' | ./k6_isomorfer > tmp_k6_unique
//


#include <iostream>

using namespace std;

// Structure holding just one K6 drawing
struct graph
{
	int   neighbours[6][5];
	bool  edgesflips[6][6];
	int   genus;
};
 
// Maximum storage for graphs - 10 woudl be sufficient too, since
// there are only 7 non-isomorf embeddings
const int MAX_STORAGE = 20;

// We use unique_graphs[0] as a storage for currently tested graph.
// the rest of array are allready found non-isomorf graphs
graph unique_graphs[MAX_STORAGE];
int graphs = 1;

// Swap two integers
void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}


// print the graph insome "standart form"
template<class T>
void print_dan_compatible(T &stream, const graph & gr)
{
	// vertices
	for (int i = 0; i < 6; i++)
		stream << i << ":" <<  gr.neighbours[i][0] <<  gr.neighbours[i][1]
			   <<  gr.neighbours[i][2]
			   <<  gr.neighbours[i][3] <<  gr.neighbours[i][4] << " ";
	
	// adjacency matrix
	for (int i = 0; i < 5; i++) 
	{
		for (int j = i+1; j < 6; j++)
			stream << gr.edgesflips[i][j];
		stream << " ";
	}
	
	stream << "  G" << gr.genus << '\n';
}

// Just a shortcut for printing graphs from unique_graphs array
template<class T>
void print_dan_compatible(T &stream, int graphID)
{
	print_dan_compatible(stream, unique_graphs[graphID]);
}

// Create edge flips matrix from int. The edgeflip matrix can be stored by bites
// in a integer. The example of a matrix and the bites:  - stadns for diagonal
//
//    -    0   0   0   0  0
//    0   -  512 256 128 64
//    0  512  -   32  16  8
//    0  256  32  -    4  2 
//    0  128  16   4   -  1
//    0   64   8   2   1  -
//

void matrix_from_int(const int &i, const int graphID)
{

unique_graphs[graphID].edgesflips[5][4] = unique_graphs[graphID].edgesflips[4][5] = i &    1;
unique_graphs[graphID].edgesflips[5][3] = unique_graphs[graphID].edgesflips[3][5] = i &    2;
unique_graphs[graphID].edgesflips[4][3] = unique_graphs[graphID].edgesflips[3][4] = i &    4;
unique_graphs[graphID].edgesflips[5][2] = unique_graphs[graphID].edgesflips[2][5] = i &    8;
unique_graphs[graphID].edgesflips[4][2] = unique_graphs[graphID].edgesflips[2][4] = i &   16;
unique_graphs[graphID].edgesflips[3][2] = unique_graphs[graphID].edgesflips[2][3] = i &   32;
unique_graphs[graphID].edgesflips[5][1] = unique_graphs[graphID].edgesflips[1][5] = i &   64;
unique_graphs[graphID].edgesflips[4][1] = unique_graphs[graphID].edgesflips[1][4] = i &  128;
unique_graphs[graphID].edgesflips[3][1] = unique_graphs[graphID].edgesflips[1][3] = i &  256;
unique_graphs[graphID].edgesflips[2][1] = unique_graphs[graphID].edgesflips[1][2] = i &  512;

}

// Load a graph from standart input
bool load_dan_compatible(int graphID)
{
	// vertices
	for (int i = 0; i < 6; i++)
	{
		int list;
		cin >> list;
		
		unique_graphs[graphID].neighbours[i][0] = (list / 10000) % 10;
		unique_graphs[graphID].neighbours[i][1] = (list / 1000)  % 10;
		unique_graphs[graphID].neighbours[i][2] = (list / 100)   % 10;
		unique_graphs[graphID].neighbours[i][3] = (list / 10)    % 10;
		unique_graphs[graphID].neighbours[i][4] = (list / 1)     % 10;
	}
	
	int flips, tmp;
	cin >> flips;
	
	// adjacency matrix
	bzero(unique_graphs[graphID].edgesflips, 6*6*sizeof(bool));

	for (int i = 4; i > 0; i--) 
	{
		cin >> tmp;
		flips = (flips << i);
		
		for (int k = 0; k < i; k++)
		{
			flips |= (tmp % 10) << k;
			tmp /= 10;
		}
	}

	cin >> unique_graphs[graphID].genus;
	
	if (!cin) return false;
		
	matrix_from_int(flips, graphID);

	return true;
}



// Try to add a new graph to the list of unique graphs
// it means store the graph and print it out
void store_new(int fromID)
{
	unique_graphs[graphs] = unique_graphs[fromID];
	graphs++;
	
	// we found a new graph so print it out :-)
	print_dan_compatible(cout, fromID);
	
	if (graphs == MAX_STORAGE)
	{
		cout << "The MAX_STORAGE it too small! :(" << endl;
		exit(1);
	}
}

// Input is graph a and output is graph tmp. The graph tmp
// is a graph a changed by mapping. Mapping just for
// bijecton [0..5] -> [0..5]. It just chages the names of
// vertices. Note that a and tmp are isomorh. 
void remap(const graph &a, graph &tmp, int mapping[])
{
	for (int i = 0;  i < 6; i++)
		for (int j = 0;  j < 5; j++)
		{
			tmp.neighbours[ mapping[i] ][j] = mapping[ a.neighbours[i][j] ];
		}
			
		
	for (int i = 0;  i < 6; i++)
		for (int j = 0;  j < 6; j++)
		{
			tmp.edgesflips[ mapping[i] ][ mapping[j] ] = a.edgesflips[i][j];
		}
}

// Inverts lists of neighbours at vertex i
void flip_list(graph &a, const int & i)
{
	// reverse orientation of the neighbour list
	swap(a.neighbours[i][0], a.neighbours[i][4]);
	swap(a.neighbours[i][1], a.neighbours[i][3]);	
}

// Flip a vertex means flip the list of neighbours and flip
// all edges containing this vertex. Flipping keeps the
// isomorphism. Using this we may always make all edges going
// from 0 to be non-flipping
void flip_vertex(graph &tmp, const int & i)
{
	// reverse orientation of the neighbour list
	flip_list(tmp, i);
	
	// change edges flips
	for (int j = 0; j < 6; j++)
	{
		tmp.edgesflips[i][j] = !tmp.edgesflips[i][j];
		tmp.edgesflips[j][i] = !tmp.edgesflips[j][i];
	}
}

// Makes all edges containing vertex 0 be a nonfliping
// I helps with isomorphism testing.
void clear_edgeflips(graph &tmp)
{
	for (int i = 1;  i < 6; i++)
		if (tmp.edgesflips[ 0 ][ i ])
		{
			flip_vertex(tmp, i);
		}
}

// Rotate the list of neighbours at vertex i. Rotating list
// does not change isomorhism since only the cylicity matters.
void rotate_list(graph &tmp, int i)
{
	int k = tmp.neighbours[i][0];
	tmp.neighbours[i][0] = tmp.neighbours[i][1];
	tmp.neighbours[i][1] = tmp.neighbours[i][2];
	tmp.neighbours[i][2] = tmp.neighbours[i][3];
	tmp.neighbours[i][3] = tmp.neighbours[i][4];
	tmp.neighbours[i][4] = k;	
}


// Make 1 first at first vertex and 0 in other vertices.
void normalize_lists(graph &tmp)
{	
	while (tmp.neighbours[0][0] != 1)
	{
		rotate_list(tmp, 0);
	}	
	
	for (int i = 1; i < 6; i++)
	{
		while (tmp.neighbours[i][0] != 0)
		{
			rotate_list(tmp, i);
		}
	}
}


// Check if the graphs are same 
bool are_same(const graph &a, const graph &b)
{
	// check neighbours
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 5; j++)
		{
			if (a.neighbours[i][j] != b.neighbours[i][j]) 
			{
				return false;
			}
		}
	}

	// check edge flips
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			if (i == j) continue;
			if (a.edgesflips[i][j] != b.edgesflips[i][j])
			{
				return false;
			}			
		}
	}

	return true;
}

// flip all neighbour lists of graph. The graph remains
// isomorhic.
void flip_graph(graph &a)
{
	for (int j = 0; j < 6; j++)
	{
		flip_list(a, j);					
	}
}

// It is determined that vertex i in graph a goes to vertex 0 in graph b.
// now we determine mathing of other vertices by scrolling
// the list at a[i] and map to other vertices. - the cyclic ordering must remain.
bool match_othervertices(const int &i, graph &a, graph &b, graph &tmp, int mapping[])
{			
	// scroll of the list of first vertex
	for (int j = 0; j < 5; j++)
	{
		// fill the mapping
		for (int k = 0; k < 5; k++)
		{
			mapping[a.neighbours[i][(k+j)%5]] = k+1;
		}	

		// remap the graph A to tmp;
		remap(a, tmp, mapping);
		
		// update tmp such that all edges from first vertex have edgesflips = 0
		clear_edgeflips(tmp);
		
		// normalize color lists such that 0 is first every time
		normalize_lists(tmp);
				
		// check it b and tmp are the same
		if (are_same(b, tmp)) return true;
	}
	
	return false;
}

// Test if unique_graphs[graphA] and unique_graphs[graphB]
// are isomorphic
bool are_isomorphic(const int graphA, const int graphB)
{
	// try map a vertex
	//A -> B
	graph &a   = unique_graphs[graphA];
	graph &b   = unique_graphs[graphB];
	graph &tmp = unique_graphs[graphs];
	
	if (a.genus != b.genus) return false;
	tmp.genus = a.genus;

	for (int i = 0; i < 6; i++)
	{
		int mapping[6];
		
		// we determine a vertex matched to zero		
		// vertex a[i] -> vertex b[0]
		mapping[i] = 0;
		
		if (match_othervertices(i, a, b, tmp, mapping)) return true;

		flip_graph(a);

		if (match_othervertices(i, a, b, tmp, mapping)) return true;
		
		flip_graph(a); // we could skip the back flip.... but...
		               // just to have consistent a at the beginning of cycle.
	}
	
	return false;
}

// Just the main :-)
int main(int argc, char **argv)
{
	while (load_dan_compatible(0))
	{

		int i;
		for (i = 1; i < graphs; i++)
		{
			if (are_isomorphic(0, i)) break;
		}
				
		if (i < graphs) continue;
		
		store_new(0);
	}
}



