//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
//  Change the neighbours and flip edges representation into faces prepresentation.
//  The program just read the input and output the faces prepresentation
//  Big part of the program is just a rip from k6_generator.cpp
//
//  Executing: 
//
//  cat tmp_k6_unique | sed 's/.://g' | sed 's/G//' | ./k6_to_face_rep > tmp_k6_faces
//

#include <iostream>
#include <vector>

using namespace std;

// there is only one - currently processed - graph in program

int   neighbours[6][5]; // List of neighbours
bool  edgesflips[6][6]; // nighbour matrix of edgeflips


// Swap integers a and b
void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}

// Used edges are for enumerating faces. We enumare them
// by walking along the cycles and check if we use an edge.
// possible valies are  
//   0  ... not used
//   1  ... one side used
//   2  ... second side used
//   3  ... both sides used
int  used_edges[6][6];

vector<int> face;

// Start walking and keep walking along an edge. We want to go
// from vertex i to vertex j and use edgeside side of the edge.
// the matrix used_edges is filled with face containing this
// edge (assuming that no edges from the face are marked)
void walk_edge(int i, int j, int edgeside)
{	
	// check if the edge was allready visited. If so just return.
	if (used_edges[i][j] & edgeside) return;

	face.push_back(i);	
	
	used_edges[i][j] |= edgeside;
	
	// flip the edge is applicable
	if (edgesflips[i][j])
	{
			if (edgeside == 1) edgeside = 2;
			else edgeside = 1;
	}

	// makr destination of edge
	if (edgeside == 1) used_edges[j][i] |= 2;
	else used_edges[j][i] |= 1;

	// find next vertex in the face
	int idx = 0;
	while (neighbours[j][idx] != i) idx++;
	
	if (edgeside == 2)
	{
		if (++idx > 4) idx = 0;
	}
	else
	{
	 	if (--idx < 0) idx = 4;
	}
		
	walk_edge(j, neighbours[j][idx], edgeside);
}


// Test if we have a wanted graph. It means we need
// to check the genus of the grap
void convert_graph()
{
	// Try to count faces.
	bzero(used_edges, sizeof(int)*36);

	int faces = 0;
	
	for (int i = 0; i < 6; i++)
	{
		for (int j = 0; j < 6; j++)
		{
			if (i == j) continue;
			
			if ((used_edges[i][j] & 1) == 0)
			{
				face.clear();
				walk_edge(i,j,1);
				// print the faces as we find them. We do not need to store them all.
				cout << "F" << faces << ": " << face.size() << "   ";
				for (unsigned int k = 0; k < face.size(); k++) cout << face[k] << " ";
				faces++;
				cout << endl;
			}
			
			if ((used_edges[i][j] & 2) == 0)
			{
				face.clear();
				walk_edge(i,j,2);
				cout << "F" << faces << ": " << face.size() << "   ";
				for (unsigned int k = 0; k < face.size(); k++) cout << face[k] << " ";
				faces++;
				cout << endl;
			}			
		}
	}
	cout << endl;
	
	// Try if genus == 2. If not, we got a fake graph. This is a secirity check.
	if ( 2 != 11 - faces)
	{
		cerr << "Graph with wrong genus!!!" << endl;
		exit(1);
	}
}


// Create edge flips matrix from int
void matrix_from_int(const int &i)
{
	
	edgesflips[5][4] = edgesflips[4][5] = i &    1;
	edgesflips[5][3] = edgesflips[3][5] = i &    2;
	edgesflips[4][3] = edgesflips[3][4] = i &    4;
	edgesflips[5][2] = edgesflips[2][5] = i &    8;
	edgesflips[4][2] = edgesflips[2][4] = i &   16;
	edgesflips[3][2] = edgesflips[2][3] = i &   32;
	edgesflips[5][1] = edgesflips[1][5] = i &   64;
	edgesflips[4][1] = edgesflips[1][4] = i &  128;
	edgesflips[3][1] = edgesflips[1][3] = i &  256;
	edgesflips[2][1] = edgesflips[1][2] = i &  512;
	
}

// Load a graph
bool load_dan_compatible()
{
	// vertices
	for (int i = 0; i < 6; i++)
	{
		int list;
		cin >> list;
		if (!cin) return false;
		
		neighbours[i][0] = (list / 10000) % 10;
		neighbours[i][1] = (list / 1000)  % 10;
		neighbours[i][2] = (list / 100)   % 10;
		neighbours[i][3] = (list / 10)    % 10;
		neighbours[i][4] = (list / 1)     % 10;
	}
	
	int flips, tmp;
	cin >> flips;
	
	// adjacency matrix
	bzero(edgesflips, 6*6*sizeof(bool));
	
	for (int i = 4; i > 0; i--) 
	{
		cin >> tmp;
		if (!cin) return false;

		flips = (flips << i);
		
		for (int k = 0; k < i; k++)
		{
			flips |= (tmp % 10) << k;
			tmp /= 10;
		}
	}
	matrix_from_int(flips);
	
	// Verify the genus
	int genus;
	cin >> genus;
	if (!cin) return false;
	
	if (genus != 2)
	{
		cerr << "The program should be use only for genus 2!" << endl;
		exit(1);
	}
	
	return true;
}

int main(int argc, char **argv)
{
	int id = 0;
	while (load_dan_compatible())
	{
		cout << "G:K6_" << id++ << " -1 v: 6 f: 9" << endl;
		convert_graph();
	}
	
	return 0;
}


