//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
// Program for generating all k-minimal graphs. All orientations are always
// clockwise. The procedure is described in Section 3. The program creates
// 4-,5-,6-,7-,8-,9- and 10-minimal graphs exclusing the trivial ones (empty circle).
//
//  Executing: 
//
//  ./kminimal > tmp_kminimal
//

#include <iostream>
#include <fstream>
#include <vector>

using namespace std;

const int MAX_VERTICES =  40;
const int MAX_FACES    =  40;
const int MAX_K        =  12;

int tested = 0;
int unique = 0;

// Cyclic version of indexing a vector
template<class T>
int at(const vector<T> &vec, int idx)
{
	int sz = vec.size();
	return vec[(idx+sz)%sz];
}

// Slightly more advanced structure graph than in k6_* programs.
struct graph
{
	int           vertices;
	int           k;          // the length ouf outer circle
	vector<int>   neighbours[MAX_VERTICES];
	
	// Does not respect any cyclic ordering
	void add_edge(int vertexA, int vertexB)
	{
		// Security chceck
		for (unsigned int i = 0; i < neighbours[vertexA].size(); i++)
			if (neighbours[vertexA][i] == vertexB) cout << "BUG" << endl;

		for (unsigned int i = 0; i < neighbours[vertexB].size(); i++)
			if (neighbours[vertexB][i] == vertexA) cout << "BUG" << endl;

		neighbours[vertexA].push_back(vertexB);
		neighbours[vertexB].push_back(vertexA);
	}
	
	// To vertexA insert vertexB to particular position:
	//  vertexA: [....vertexA_prev, vertexA_next, ....]   ->  [....vertexA_prev, vertexB, vertexA_next, ....]
	void insert_neighbour(int vertexA_prev, int vertexA, int vertexA_next, int vertexB)
	{
		vector<int> neigh;
		vector<int> &original = neighbours[vertexA];
		for (int i = 0; i < (signed int)original.size(); i++)
		{
			if (original[i] == vertexA_prev)
			{
				/*
				if (at(original, i-1) == vertexA_next)
				{
					neigh.push_back(vertexB);
					neigh.push_back(vertexA_prev);
					continue;
				}
				*/
				if (at(original, i+1) == vertexA_next)
				{
					neigh.push_back(vertexA_prev);
					neigh.push_back(vertexB);
					continue;
				}				
			}
			neigh.push_back(original[i]);
		}
		if (original.size() == neigh.size())
		{
			cerr << "Inserting failed!!!!" << endl;
			exit(1);
		}
		original = neigh;
	}
	
	// Adding adge 
	void add_edge_circular(int vertexA_prev, int vertexA, int vertexA_next, int vertexB_prev, int vertexB, int vertexB_next)
	{
		// Security chceck
		for (unsigned int i = 0; i < neighbours[vertexA].size(); i++)
			if (neighbours[vertexA][i] == vertexB) cout << "BUG" << endl;
		
		for (unsigned int i = 0; i < neighbours[vertexB].size(); i++)
			if (neighbours[vertexB][i] == vertexA) cout << "BUG" << endl;
		
		insert_neighbour(vertexA_prev, vertexA, vertexA_next, vertexB);
		insert_neighbour(vertexB_prev, vertexB, vertexB_next, vertexA);
	}
		
	int add_vertex()
	{
		return vertices++;
	}
	
	// Removes an edge
	void remove_edge(int vertexA, int vertexB)
	{
		for (int i = 0; i < (signed int)neighbours[vertexA].size(); i++)
		{
			if (neighbours[vertexA][i] == vertexB)
			{
				if ( i != (signed int)neighbours[vertexA].size() - 1)
				{
					neighbours[vertexA][i] = neighbours[vertexA][neighbours[vertexA].size() - 1];
				}
				neighbours[vertexA].pop_back();
			}
		}

		for (int i = 0; i < (signed int)neighbours[vertexB].size(); i++)
		{
			if (neighbours[vertexB][i] == vertexA)
			{
				if ( i != (signed int)neighbours[vertexB].size() - 1)
				{
					neighbours[vertexB][i] = neighbours[vertexB][neighbours[vertexB].size() - 1];
				}
				neighbours[vertexB].pop_back();
			}
		}
	}
	
	void erase()
	{
		for (int i = 0; i < MAX_VERTICES; i++)
		{
			neighbours[i].clear();
		}
		vertices = 0;
	}
	
	graph()
	{
			k = 0;
			vertices = 0;
	}
};
 

typedef vector<int>  face; // index of vertices going clockwise

struct faces_queue
{
	face faces[100];
	int  min_valid;
	int  max_valid;
	
	faces_queue()
	{
			min_valid = 0;
			max_valid = 0;
	}
	
	void add_face(const face &f)
	{
		if (f.size() <= 3) return;
		faces[max_valid++] = f;
	}
};

vector<graph> unique_graphs[MAX_K];
vector<graph> tested_graphs;


void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}


template<class T>
void print_dan_compatible(T &stream, const graph & g)
{
	stream << g.k << " v: " << g.vertices << endl;
	
	// vertices
	for (int i = 0; i < g.vertices; i++)
	{
		stream << i << ": " << g.neighbours[i].size() << "   ";
		for (unsigned int j = 0; j < g.neighbours[i].size(); j++)
		{
			stream << g.neighbours[i][j] << " ";
		}
		stream << endl;
	}
	stream << endl;
}

template<class T>
bool load_dan_compatible(T &stream, graph & g)
{
	g.erase();
	
//	cout << "v: " << g.vertices << endl;
	
	stream >> g.k;

	if (!stream) return false;
	
	int vertices;
	stream >> vertices;
	
	if (!stream) return false;

	for (int i = 0; i < vertices; i++)
	{
		g.add_vertex();
	}
	
	for (int i = 0; i < vertices; i++)
	{
		int degree;
		stream >> degree;
		
		if (!stream)
		{
			cerr << "Something wrong while loading vertex " << i << endl;
			return false;
		}

		
		for (int j = 0; j < degree; j++)
		{
			int e;
			stream >> e;

			if (!stream)
			{
				cerr << "Something wrong with loading neighbour " << j << " at neighbour " << i << endl;
				return false;
			}

			if (e == i)
			{
				cerr << "Loop at vertex " << i << endl;
				return false;
			}
			
			if (e > i)
			{
			//	cout << e << "-" << i << endl;
			// this breaks the cyclic order!!
				g.add_edge(e, i);
			}
		}
	}
	
	return true;
}

/////////////////////////////////////////////////////////////// Isomorphism testing

// Return false if graphs amb have different degrees/vertices
// a quick check ro say false to isomorphism
bool check_degrees(const graph &a, const graph &b)
{
	if (a.k != b.k)
	{
	//	cout << "Warinng - different l" << endl;
		return false;
	}
	
	if (a.vertices != b.vertices) return false;
	
	for (int deg = 1; deg < a.vertices - 1; deg++)
	{
		int cnta = 0;
		for (int i = 0; i < a.vertices; i++)
			if ((signed int)a.neighbours[i].size() == deg) cnta++;
		
		int cntb = 0;
		for (int i = 0; i < b.vertices; i++)
			if ((signed int)b.neighbours[i].size() == deg) cntb++;
		
		if (cnta != cntb) return false;
	}
	
	return true;
}


int map_atob[MAX_VERTICES];
int map_btoa[MAX_VERTICES];

bool can_map(const graph &a, const graph &b, int i, int j)
{
	if (map_btoa[j] != -1) return false;
	if (a.neighbours[i].size() !=  b.neighbours[j].size()) return false;
	
	// map only vertices from outer circle to outer circle
	if (i <  a.k && j >= b.k) return false;
	if (i >= a.k && j <  b.k) return false;
	
	
	// test if all previous neighbours are mapped correctly
	for (unsigned int vertA = 0; vertA < a.neighbours[i].size(); vertA++)
	{
		if (a.neighbours[i][vertA] > i) continue;
		int mapped = map_atob[a.neighbours[i][vertA]];
		
		unsigned int vertB;
		for (vertB = 0; vertB < b.neighbours[j].size(); vertB++)
			if (b.neighbours[j][vertB] == mapped) break;
		
		// we found previously mapped vertex adjacent to i but it's
		// image is not mapped as a neigbbour of j.
		if (vertB == b.neighbours[j].size()) return false;
	}
	
	return true;
}

bool match_othervertices(const int i, const graph &a, const graph &b)
{	
	if (i >= a.vertices) return true;
	
	// scroll of the list of first vertex
	for (int j = 0; j < a.vertices; j++)
	{
		if (!can_map(a, b, i, j)) continue;
		
		// fill the mapping
		map_atob[i] = j;
		map_btoa[j] = i;
		
		if (match_othervertices(i+1, a, b)) return true;
		
		map_btoa[j] = -1;
	}
	
	return false;
}

bool are_isomorphic(const graph &a, const graph &b)
{
	// try map a vertex
	//A -> B
	if (a.k != b.k)
	{
		cerr << "internal error - comparing graphs with different outer circle" << endl;
		exit(1);
	}
		
	if (!check_degrees(a, b)) 
	{
		return false;
	}
		
	for (int i = 0; i < a.vertices; i++)
	{
		map_atob[i] = -1;	
		map_btoa[i] = -1;	
	}
	
	return match_othervertices(0, a, b);
}

/////////////////////////////////////////////////////////////// Coloring testing


int coloring[MAX_VERTICES];


// Check if vertex color is not confligting with other
// previous vertices. true == all is allright
bool check_vertex(const graph &g, const int vertexID)
{
	for (unsigned int testedID = 0; testedID < g.neighbours[vertexID].size(); testedID++)
		if (coloring[vertexID] == coloring[g.neighbours[vertexID][testedID]]
			&& g.neighbours[vertexID][testedID] < vertexID) return false;
	
	return true;
}

bool extend_precoloring(const graph &g, const int i)
{
	if (i >= g.vertices) 
	{
		return true;
	}
	
	for (int col = 0; col < 5; col++)
	{
		coloring[i] = col;
		
		// some chord...
		if (!check_vertex(g, i)) continue; 
		
		if (extend_precoloring(g, i+1)) return true;
	}
	
	return false;
}

bool precolor_vertex(const graph &g, const int i)
{
	if (i >= g.vertices) 
	{
		// OK, there is nothing left to color
		return true;
	}
		
	// "Precoloring of cycle"
	if ( i == 0 )
	{
		// We can allways assume that the first vertex gets color 0 and
		// second gets color 1 (they are connected since on a circle)
		coloring[0] = 0;
		coloring[1] = 1;
		
		// recursive
		return precolor_vertex(g, 2);
		
	}
	else if ( i < g.k)
	{
			for (int col = 0; col < 5; col++)
			{
				if (coloring[i - 1] == col) continue;
				if ( i == g.k-1 && coloring[0] == col) continue;
	
				coloring[i] = col;
				
				// some chord...
				if (!check_vertex(g, i)) return false; 
				
				// recursive
				if (!precolor_vertex(g, i+1)) return false;
			}
		return true;
	}
	
	// Coloring of inner vertices
	else
	{
		return extend_precoloring(g, i);
	}
	
//	return false;
}
						
// h is missong one edge
bool is_edge_critical(const graph &g, const graph &h, const int i)
{
	if (i >= g.vertices) 
	{
		return false;
	}
	
	// "Precoloring of cycle"
	if ( i == 0 )
	{
		coloring[0] = 0;
		coloring[1] = 1;
		
		// recursive
		return is_edge_critical(g, h, 2);
		
	}
	else if ( i < g.k)
	{
		for (int col = 0; col < 5; col++)
		{
			if (coloring[i - 1] == col) continue;
			if ( i == g.k-1 && coloring[0] == col) continue;
			
			coloring[i] = col;
			
			// some chord...
			if (!check_vertex(g, i))
			{
		//		cout << " check " << i << " " << check_vertex(h, i) << endl;
				if (check_vertex(h, i)) return true;
			}
				
			// recursive
			if (is_edge_critical(g, h, i+1)) return true;
		}
		return false;
	}
	
	// Coloring of inner vertices
	else
	{
		
		if (!extend_precoloring(g, i) && extend_precoloring(h, i)) 
		{
			return true;
		}
		
		return false;	
	}
//	return false;
}


bool is_removeable(int vertexA, int vertexB, int k)
{
	if (vertexA >= k || vertexB >= k) return true;
	if (vertexA - vertexB == 1) return false;
	if (vertexB - vertexA == 1) return false;
	if (vertexA == 0 && vertexB == k-1) return false;
	if (vertexB == 0 && vertexA == k-1) return false;
	
	return true;
}

// Try to add new graph
// return true if there is a chance of getting a new graph
bool try_graph(const graph &g)
{
	
	if (tested++ % 5000 == 0) 
	{
		cerr << "Tested: " << tested << " Unique tests: " << tested_graphs.size() << " Found: " << unique_graphs[g.k].size() << endl;
	}
	

	// check for allready tested graphs
	for (unsigned int i = 0; i < tested_graphs.size(); i++)
	{
		if (are_isomorphic(tested_graphs[i], g)) return false;
	}
	tested_graphs.push_back(g);
	
		
	// 5-pre coloring test - if all precolorings are
	// extensible we are not interested in such a graph
	// but it could be extended...
	if (precolor_vertex(g, 0)) 
	{		
		//cerr << "Precoloring failed" << endl;		
		return true;
	}
	
	
	// Test if it is l-superminimal
	for (int v = 0; v < g.vertices; v++)
	{
		const vector<int>  &neighbours = g.neighbours[v];
		
		for (unsigned int i = 0; i < neighbours.size(); i++)
		{
			if ( neighbours[i] < v && is_removeable(v, neighbours[i], g.k) )
			{
				graph h = g;
				h.remove_edge(v, neighbours[i]);
				
				if (!is_edge_critical(g, h, 0)) 
				{
				//	cerr << "Falese witness - not critical" << endl;
					return true;
				}
			}
		}
	}
	
	
	// Test if it is not isomorphic to other graphs
	for (unsigned int gID = 0; gID < unique_graphs[g.k].size(); gID++)
	{
		if (are_isomorphic(g, unique_graphs[g.k][gID])) 
		{
			//cerr << "Isomorf one" << endl;
			return false;
		}
	}
	
	// Cool - we have a new graph :-)
	cout << "G:" << unique_graphs[g.k].size() << " ";
	print_dan_compatible(cout, g);

	// add it
	unique_graphs[g.k].push_back(g);

	return true;
}

/////////////////////////////////////////////////////////////// Initial stuff

// For debugging
void print_queue(const faces_queue &q)
{
	cerr << "{";
	int i;
	for (i = 0; i < q.max_valid; i++)
	{
		for (unsigned int j = 0; j < q.faces[i].size(); j++)
			cerr << " " << q.faces[i][j];
		cerr << ",";
	}
	cerr << " }" << endl;
	
}

// For debugging
void print_face(const face &f)
{
	cerr << "[";
	for (unsigned int j = 0; j < f.size(); j++)
		cerr << " " << f[j];
	cerr << "]" << endl;
	
}

// Create a graph of outer circle.
void create_circle(graph &g, faces_queue &q, int k)
{
	// Generate the graph
	g.vertices = k;
	g.k        = k;
	
	
	// We can use destroying add_edge version since the vertices
	// will have ony 2 neighbours - thus allways clockwise :-)
	for (int i = 0; i < k-1; i++)
	{
		g.add_edge(i,i+1);		
	}
	g.add_edge(0,k-1);

	
	// Init the faces queue
	q.min_valid = 0;
	q.max_valid = 1;
	
	q.faces[0].clear();
	for (int i = 0; i < k; i++)
	{
		q.faces[0].push_back(i);
	}
}

/////////////////////////////////////////////////////////////// Faces subdivision


void divide_faces(const graph &g, faces_queue &q, int depth, bool try_new = true)
{
	if (try_new && !try_graph(g)) return;
		
	if (q.min_valid >= q.max_valid) return;
		
	int min =  q.min_valid;  // current size of queue
	int max =  q.max_valid;
	
	face &f = q.faces[q.min_valid];
	
	q.min_valid++;

	if (f.size() < 3) // faces of size < 3 are obvioisly an error
	{
		cerr << "small face!!!!" << endl;
		exit(1);
	}
	
	// just keep going... It may be a good idea to 
	divide_faces(g, q, depth+1, false);
	
	if (f.size() > 3)
	{
		
		if ((signed int)f.size() == g.k && depth < 1)
		{
			// Divide face and try...
			/// Add chord
			//for (unsigned int i = 0; i < f.size(); i++)
			{
				unsigned int i = 0;
				//for (unsigned int j = i+2; j < f.size(); j++)
				for (unsigned int j = i+2; j < 1+f.size()/2; j++)
					{
					//	if (!can_be_chord(i, j, f.size())) continue;
					if (i == 0 && j == f.size()-1) continue;
				
				
					graph h = g;
				
					//	cerr << "adding chord " << f[i] << f[j] << endl;
				
					// the face is clockwise and we want to keep vertices
					// also clockwise, so we need to revers
					h.add_edge_circular(at(f, i+1), f[i], at(f, i-1),
										at(f, j+1), f[j], at(f, j-1));
				
					face a,b; // two new faces are created when splitting by a chord

					for (unsigned int x = 0; x < f.size(); x++)
					{
						if (x <= i || x >= j) a.push_back(f[x]);
						if (x >= i && x <= j) b.push_back(f[x]);
					}
				
					q.max_valid = max;
				
					q.add_face(a);
					q.add_face(b);
		
					// continue deeper
					divide_faces(h, q, depth+1);
				}
			} // enc of adding a chord
		
			/// Add inner vertex of degree 3
			// The vertex has 3 edges -> 3 for cycles
			//for (int i = 0; i < (signed int)f.size(); i++)
			int i = 0;
			{
				for (int j = i+1; j < (signed int)f.size(); j++)
				{
					for (int k = j+1; k < (signed int)f.size(); k++)
					{
						graph h = g;
					
						int v = h.add_vertex();
					
						//cerr << "Adding " << f[i] << f[j] << f[k] << v << endl;
						
						// Insert edges of vertex to graph
						h.insert_neighbour(at(f, i+1), f[i], at(f,i-1), v);
						h.insert_neighbour(at(f, j+1), f[j], at(f,j-1), v);
						h.insert_neighbour(at(f, k+1), f[k], at(f,k-1), v);
						h.neighbours[v].push_back(f[i]);
						h.neighbours[v].push_back(f[j]);
						h.neighbours[v].push_back(f[k]);
					
						// Create new faces - 3 faces
						face a,b,c;
						for (int x = 0; x < (signed int)f.size(); x++)
						{
							if (x <= i || x >= k) a.push_back(f[x]);
							if (x >= i && x <= j) b.push_back(f[x]);
							if (x >= j && x <= k) c.push_back(f[x]);
							if (x == i) a.push_back(v);
						}
						b.push_back(v);
						c.push_back(v);
					
						q.max_valid = max;
					
						q.add_face(a);
						q.add_face(b);
						q.add_face(c);
					
						// Continue deeper
						divide_faces(h, q, depth+1);					
					}
				}
			}
		}
		// end of adding vertex of size 3
		
		// Glue inside something allready done...
		// We gule allready created critical graphs
		else
		{
			int sz     = f.size(); // This is size of current face
			for (unsigned int fcID = 0; fcID < unique_graphs[sz].size(); fcID++)
			{
				//		const graph &topaste = unique_graphs[sz][fcID];
				int mapping[unique_graphs[sz][fcID].vertices];
				
				for (int flip = 0; flip < 2; flip++)
				{
					//cerr << "flip glue: " << flip << endl;
			
					for (int shift = 0; shift < sz; shift++) // rotating the graph
					{
						graph h = g;
						const graph &topaste = unique_graphs[sz][fcID]; // we need to refresh the reference
				
						// outer circular vertices
						for (int i = 0; i < sz; i++)
						{
							if (flip) mapping[i] = f[(-i+shift+sz)%sz];	
							else      mapping[i] = f[(i+shift)%sz];							
						}
					
						// inner vertices - create copies from pastev graph
						for (int i = sz; i < topaste.vertices; i++)
						{
							mapping[i] = h.add_vertex();
						}
						
						// map neighbours for inner vertices
						for (int i = sz; i < topaste.vertices; i++)
						{
							for (unsigned int j = 0; j < topaste.neighbours[i].size(); j++)
							{
								if (!flip)
								{
									h.neighbours[mapping[i]].push_back(mapping[topaste.neighbours[i][j]]);
								} 
								else 
								{
									// reverse if flipping
									h.neighbours[mapping[i]].push_back(mapping[topaste.neighbours[i][topaste.neighbours[i].size() - j - 1]]);
								}
							}
						}
							 
						// neighbours of the outer vertices... will bee cooool piece of program.....
						for (int i = 0; i < sz; i++)
						{
							const vector<int> &tomap = topaste.neighbours[i];
							
							// only 2 neighbours -> no need to modify anything
							if (tomap.size() == 2) continue;
							
							vector<int> neigh; // the new list
							vector<int> &original = h.neighbours[mapping[i]];
							
							int prev  = (i+sz-1)%sz;
							int next  = (i+sz+1)%sz;
							
							int prevM, nextM;
							if (!flip)
							{
								prevM = mapping[prev]; 
								nextM = mapping[next];
							}
							else
							{
								prevM = mapping[next]; 
								nextM = mapping[prev];								
							}
							
							// note that at vertex mapping we need next to be before prev - in a
							// cyclic order :)
							
							// we create a new ordering...
							//cerr << "sz = " << sz << " prev=" << prev << " next=" << next << endl;
							for (int j = 0; j < (signed int)original.size(); j++)
							{
								if (original[j] == prevM)
								{
									// This is natural mapping
									if (at(original, j-1) == nextM && !flip)
									{
									//	cerr << "not flip" << endl;
										int added = 0;										
										int k = 0;
										while(tomap[k] != next) k++;
										for ( k = (k+1)%tomap.size(); tomap[k] != prev; k = (k+1)%tomap.size())
										{
											neigh.push_back(mapping[tomap[k]]);
											added++;
										}
										if (added == 0)
										{
											cerr << "internal failure!" << endl;
											exit(1);
										}										
										neigh.push_back(prevM);
										continue;
									}
									
									// this is flipped mapping
									if (at(original, j-1) == nextM && flip)
									{
									//	cerr << "flip" << endl;
										int added = 0;										
										int k = tomap.size()-1;
										while(tomap[k] != prev) k--;
										for ( k = (k+tomap.size()-1)%tomap.size(); tomap[k] != next; k = (k+tomap.size()-1)%tomap.size())
										{
											neigh.push_back(mapping[tomap[k]]);
											added++;
										}
										if (added == 0)
										{
											cerr << "internal failure!" << endl;
											exit(1);
										}
										neigh.push_back(prevM);					
										continue;
									}
									
									cerr << "this is something wrong!!!" << endl;
									cerr << "sz=" << sz << " prev=" << prev << " next=" << next << endl;
									cerr << "prevM=" << prevM << " nextM=" << nextM << endl;
									cerr << "i = " << i << " j = " << j << " flip = " << flip << endl;
									cerr << "mapping[i]=" << mapping[i] << endl;
									print_dan_compatible(cerr, topaste);
									print_dan_compatible(cerr, h);
									print_face(f);
									
									exit(1);
									
								}
								neigh.push_back(original[j]);
							}
							
							if (original.size() == neigh.size()) // check if we added something!
							{
								cerr << "No neighbour added!" << endl;
								exit(1);
							}
							
							original = neigh;
							
						}
				
						q.max_valid = max;
				
						divide_faces(h, q, depth);
					}
				}
			}
			// End of gluing		
		}
	}
	q.max_valid = max;
	q.min_valid = min;
}

/////////////////////////////////////////////////////////////// Main

template<class T>
void preload_graphs(T &stream, bool test)
{
	int loaded = 0;
	
	graph g;
	while(load_dan_compatible(stream, g))
	{
		if ((++loaded)%10 == 0)
		{
			cerr << "Loaded: " << loaded << endl;
		}
		
		if (test)
		{
			tested_graphs.clear();
			cout << try_graph(g) << endl;
		}
		else
		{
			unique_graphs[g.k].push_back(g);
		}
		
/*		{
			cout << "Invalid input!!" << endl;
			print_dan_compatible(cout, g);
			exit(1);
		}
*/
	}
	
	cerr << "preloaded graphs" << endl;
}

int main(int argc, char **argv)
{
	// fill with initial k == 3, 4
	for (int k = 4; k < 11;  k++)
	{
		tested = 0;
		int round = 0;
		int unique_for_k = -1;
		
		graph       g;
		faces_queue q;
		create_circle(g, q, k);
		
		cerr << endl << " ------------ " << k << " ----------- " << endl;
		
		while (unique_for_k != (signed int) unique_graphs[k].size())
		{
			tested = 0;
			tested_graphs.clear();

			cerr << "round: " << round++ << endl;

			unique_for_k = unique_graphs[k].size();
			divide_faces(g, q, 0);
		}
		cerr << "Unique graphs of size " << k << ": " <<  unique_graphs[k].size() << endl;
	}

}


