//
//  project:  Generating all 6-critical graphs on the Klein bottle
//  authors:  Ken-ichi Kawarabayashi, Daniel Kral, Jan Kyncl, Bernard Lidicky
//  www:      http://kam.mff.cuni.cz/~bernard/klein
//  
//
// Program for changing representation of k-minimal graphs from vertex and 
// cyclic order to face representation. Every face is represented as a circle
// but not necessary with same orientation
//
//  Executing: 
//
// cat tmp_kminimal | sed 's/[^ ]*:[^ ]*//g' | ./kminimal_to_faces > tmp_kminimal_faces
//

 
#include <iostream>
#include <vector>

const int MAX_K        =  12;

using namespace std;

typedef vector<int>  face; // index of vertices going clockwise


struct graph
{
	int           vertices;
	int           k;
	
	vector<face> faces; 	
	
	bool                  neigh_valid;
	vector< vector<int> >   neighbours;
	
	
	// Only vertex representation
	void add_edge(int vertexA, int vertexB)
	{
		// Security chceck
		for (unsigned int i = 0; i < neighbours[vertexA].size(); i++)
			if (neighbours[vertexA][i] == vertexB) cout << "BUG" << endl;
		
		for (unsigned int i = 0; i < neighbours[vertexB].size(); i++)
			if (neighbours[vertexB][i] == vertexA) cout << "BUG" << endl;
		
		neighbours[vertexA].push_back(vertexB);
		neighbours[vertexB].push_back(vertexA);
	}
	
	int add_vertex()
	{
		neighbours.resize(vertices+1);
		return vertices++;
	}
	
	void set_v_f(int v, int f)
	{
		neighbours.resize(v);
		faces.resize(f);
		vertices = v;
	}

	bool is_edge(int i, int j)
	{
		for (unsigned int k = 0; k < neighbours[i].size(); k++)
		{
			if (neighbours[i][k] == j) return true;
		}
		return false;
	}
	
	void erase()
	{
		neighbours.clear();
		vertices = 0;
		neigh_valid = false;
		faces.clear();
	}
	
	graph()
	{
		k = 0;
		vertices = 0;
		neigh_valid = false;
	}
};

void swap(int &a, int &b)
{
	int tmp = a;
	a = b;
	b = tmp;
}


template<class T>
void print_dan_compatible(T &stream, const graph & g)
{
	stream << g.k << " v: " << g.vertices << " f: " << g.faces.size() << endl;
	
	// faces
	for (unsigned int i = 0; i < g.faces.size(); i++)
	{
		stream << i << ": " << g.faces[i].size() << "   ";
		for (unsigned int j = 0; j < g.faces[i].size(); j++)
		{
			stream << g.faces[i][j] << " ";
		}
		stream << endl;
	}
	stream << endl;
}

template<class T>
void print_vertices(T &stream, const graph & g)
{
	stream << g.k << " v: " << g.vertices << " f: " << g.faces.size() << endl;
	
	// faces
	for (int i = 0; i < g.vertices; i++)
	{
		stream << i << ": " << g.neighbours[i].size() << "   ";
		for (unsigned int j = 0; j < g.neighbours[i].size(); j++)
		{
			stream << g.neighbours[i][j] << " ";
		}
		stream << endl;
	}
	stream << endl;
}


template<class T>
bool load_neighbours_list(T &stream, graph & g)
{
	g.erase();
	
	//	cout << "v: " << g.vertices << endl;
	
	stream >> g.k;
	
	if (g.k >= MAX_K || g.k < 4)
	{
		cerr << "Found graph with k == " << g.k << ". Only valid is 3 < " << g.k << " < " << MAX_K << endl;
	}

	
	if (!stream) return false;
	
	int vertices;
	stream >> vertices;
	
	if (!stream) return false;
	
	g.set_v_f(vertices, 0);
	
	for (int i = 0; i < vertices; i++)
	{
		int degree;
		stream >> degree;
		
		if (!stream)
		{
			cerr << "Something wrong while loading vertex " << i << endl;
			return false;
		}
		
		
		for (int j = 0; j < degree; j++)
		{
			int e;
			stream >> e;
			
			if (!stream)
			{
				cerr << "Something wrong with loading neighbour " << j << " at neighbour " << i << endl;
				return false;
			}
			
			if (e == i)
			{
				cerr << "Loop at vertex " << i << endl;
				return false;
			}
			
			g.neighbours[i].push_back(e);
		}
	}
	
	g.neigh_valid = true;
	
	return true;
}


int  used_edges[100][100];

void walk_edge(int i, int j, int edgeside, graph &g, face &f)
{	
	//	cout << "i=" << i << ", j=" << j;
	//  cout << i << ","<< j << ", " << edgeside << " -> ";
	
	
	if (used_edges[i][j] & edgeside) 
	{
		//cout << "end" << endl;
		return;
	}
	
	used_edges[i][j] |= edgeside;
	
		
	if (edgeside == 1) used_edges[j][i] |= 2;
	else used_edges[j][i] |= 1;
	
	
	int idx = 0;
	while (g.neighbours[j][idx] != i) idx++;
	
	// now vertex is j
	if (edgeside == 2)
	{
		if (++idx >= (signed int)g.neighbours[j].size()) idx = 0;
	}
	else
	{
		if (--idx < 0) idx = g.neighbours[j].size()-1;
	}
	
	f.push_back(g.neighbours[j][idx]);
	
	// cout << idx << ":" << g.neighbours[j][idx] << " " << endl;
	
	
	walk_edge(j, g.neighbours[j][idx], edgeside, g, f );
}


void create_face_rep(graph &g)
{
	// Try to count faces.
	bzero(used_edges, sizeof(int)*100*100);
	

	face f;

	// thow out the outer face
	walk_edge(0, 1, 2, g, f);
	
	for (int i = 0; i < g.vertices; i++)
	{
		for (int l = 0; l < (signed int) g.neighbours[i].size(); l++)
		{
			int j = g.neighbours[i][l];
			
			if ((used_edges[i][j] & 1) == 0)
			{
				f.clear();
				walk_edge(i,j,1, g, f);
				g.faces.push_back(f);
				//				cout << endl;
			}
			
			if ((used_edges[i][j] & 2) == 0)
			{
				f.clear();
				walk_edge(i,j,2, g, f);
				g.faces.push_back(f);
				//				cout << endl;
			}
			
		}
	}
}



int main(int argc, char **argv)
{
	graph g;

	int cnt[MAX_K];
	for (int i = 0; i < MAX_K; i++) cnt[i] = 0;
	
	while (load_neighbours_list(cin, g))
	{
//		print_vertices(cout, g);		
//		cout << "creating vert rep" << endl;
		
		create_face_rep(g);
		
		cout << "G:" << cnt[g.k]++ << " ";
		
		print_dan_compatible(cout, g);
	}	
}


